/*
 * =====================================================================================
 * 
 *        Filename:  applet.c
 * 
 *     Description:  main of the Gnome VM Applet
 * 
 *         Version:  $Id: applet.c,v 1.8 2006/07/14 14:21:50 kzak Exp $
 *         Created:  14.12.2005 09:07:36 CET
 * 
 *          Author:  Karel Zak <kzak@redhat.com>
 * 
 * =====================================================================================
 */

#include <stdio.h>
#include <stdlib.h>
#include <config.h>
#include <string.h>
#include <sys/types.h>
#include <sys/wait.h>

#include <gtk/gtk.h>
#include <libgnome/libgnome.h>
#include <libgnomeui/gnome-help.h>
#include <panel-applet.h>
#include <panel-applet-gconf.h>
#include <gconf/gconf-client.h>
#include <dbus/dbus-glib.h>

#include "applet.h"


static void vm_about_cb(BonoboUIComponent *, VmAppletCtl *, const char * );
static void vm_help_cb(BonoboUIComponent *, VmAppletCtl *, const char * );
static void vm_prop_cb(BonoboUIComponent *uic, VmAppletCtl *ctl, const char *verb );

static void vm_ctl_destroy(VmAppletCtl *ctl);
static void vm_destroy_cb(GtkWidget *widget, VmAppletCtl *ctl);

static GtkWidget *vm_mitem_os(VmAppletDomain *domain, VmAppletMenuDesc *desc, GtkWidget *item);
static GtkWidget *vm_mitem_state(VmAppletDomain *domain, VmAppletMenuDesc *desc, GtkWidget *item);
static GtkWidget *vm_mitem_activity(VmAppletDomain *domain, VmAppletMenuDesc *desc, GtkWidget *item);
static GtkWidget *vm_mitem_id(VmAppletDomain *domain, VmAppletMenuDesc *desc, GtkWidget *item);
static GtkWidget *vm_mitem_name(VmAppletDomain *domain, VmAppletMenuDesc *desc, GtkWidget *item);
static GtkWidget *vm_mitem_maxmem(VmAppletDomain *domain, VmAppletMenuDesc *desc, GtkWidget *item);
static GtkWidget *vm_mitem_usemem(VmAppletDomain *domain, VmAppletMenuDesc *desc, GtkWidget *item);
static GtkWidget *vm_mitem_ncpus(VmAppletDomain *domain, VmAppletMenuDesc *desc, GtkWidget *item);
static GtkWidget *vm_mitem_cputime(VmAppletDomain *domain, VmAppletMenuDesc *desc, GtkWidget *item);
static GtkWidget *vm_mitem_uuid(VmAppletDomain *domain, VmAppletMenuDesc *desc, GtkWidget *item);

#ifdef ENABLE_CONSOLEHELPER
static void vm_amenu_suspend_cb(GtkMenuItem *menuitem, gpointer data);
static void vm_amenu_resume_cb(GtkMenuItem *menuitem, gpointer data);
static void vm_amenu_destroy_cb(GtkMenuItem *menuitem, gpointer data);
static void vm_amenu_shutdown_cb(GtkMenuItem *menuitem, gpointer data);
static void vm_amenu_terminal_cb(GtkMenuItem *menuitem, gpointer data);
#endif
static void vm_amenu_details_cb(GtkMenuItem *menuitem, gpointer data);
static void vm_amenu_console_cb(GtkMenuItem *menuitem, gpointer data);
static void vm_amenu_edit_cb(GtkMenuItem *menuitem, gpointer data);

static void vm_popmenu_update(VmAppletCtl *ctl);
static gint vm_update(VmAppletCtl *ctl);
static void vm_icon_update(VmAppletCtl *ctl, int sz, gboolean refresh);
static void vm_domain_free(VmAppletDomain *dom);
static char *vm_domainstate_to_string(virDomainState state);

static VmAppletDomain *vm_domainslist_rem(VmAppletDomain *list, VmAppletDomain *dom);
static VmAppletDomain *vm_domainslist_add(VmAppletDomain *list, VmAppletDomain *dom);
static VmAppletDomain *vm_domainslist_lookup(VmAppletDomain *list, char *uuid);
static void vm_domainslist_free(VmAppletDomain *list);

static gboolean vm_mngr_method_call(VmAppletCtl *ctl, const char *methodname,
                        VmAppletDomain *dom, const char *uri);

static int vm_domains_update_from_conn(VmAppletCtl *ctl, VmAppletConn *conn);

static int instances;

static const BonoboUIVerb vm_menu_verbs [] = 
{
    BONOBO_UI_UNSAFE_VERB ("VmAppletProp",  vm_prop_cb),
	BONOBO_UI_UNSAFE_VERB ("VmAppletHelp",  vm_help_cb),
	BONOBO_UI_UNSAFE_VERB ("VmAppletAbout", vm_about_cb),
    BONOBO_UI_VERB_END
};

/* Applet Icons */
static const VmAppletStockIcon vm_stock_icons[] =
{
    { VM_STOCK_BLOCKED,     VM_ICONDIR "/vm-blocked.png"     },  
    { VM_STOCK_PAUSED,      VM_ICONDIR "/vm-paused.png"      },
    { VM_STOCK_RUNNING,     VM_ICONDIR "/vm-running.png"     },  
    { VM_STOCK_SHUTDOWN,    VM_ICONDIR "/vm-shutdown.png"    }, 
    { VM_STOCK_SHUTOFF,     VM_ICONDIR "/vm-shutoff.png"     },
    { VM_STOCK_UNKNOWN,     VM_ICONDIR "/vm-unknown.png"     },
    { VM_STOCK_DO_DESTROY,  VM_ICONDIR "/vm-do-destroy.png"  },
    { VM_STOCK_DO_SUSPEND,  VM_ICONDIR "/vm-do-suspend.png"  },
    { VM_STOCK_DO_SHUTDOWN, VM_ICONDIR "/vm-do-shutdown.png" },
    { VM_STOCK_DO_RESUME,   VM_ICONDIR "/vm-do-resume.png"   },
    { VM_STOCK_DO_TERMINAL,  VM_ICONDIR "/vm-do-terminal.png"  }
};

#define VM_ICON  (VM_ICONDIR "/vm-applet.png")
#define VM_XICON  (VM_ICONDIR "/vm-xapplet.png")

/* Applet domains menu fields */
static VmAppletMenuDesc vm_menu_desc[] =
{
    /* data Id, title, gconf, callback, enable */
    { VM_DATA_ACTIVITY, "Activity icon",    "mitem_activity", vm_mitem_activity, TRUE  },
    { VM_DATA_ID,       "Id",               "mitem_id",     vm_mitem_id,         TRUE  },
    { VM_DATA_NAME,     "Name",             "mitem_name",   vm_mitem_name,       TRUE  },

/* Xen reports strange numbers for MaxMemory, so we don't 
 * present this junk to end-users.  --kzak 03/2006 
 *  { VM_DATA_MAXMEM,   "Max memory",       "mitem_maxmem", vm_mitem_maxmem,     FALSE },
 */   
    { VM_DATA_USEDMEM,  "Used memory",      "mitem_usemem", vm_mitem_usemem,     TRUE  },
    { VM_DATA_NCPUS,    "Number of CPUs",   "mitem_ncpus",  vm_mitem_ncpus,      FALSE },
    { VM_DATA_OSTYPE,   "OS type",          "mitem_os",     vm_mitem_os,         FALSE },
    { VM_DATA_STATE,    "State",            "mitem_state",  vm_mitem_state,      FALSE },
    { VM_DATA_CPUTIME,  "CPU Time",         "mitem_cputime",vm_mitem_cputime,    FALSE },
    { VM_DATA_UUID,     "UUID",             "mitem_uuid",   vm_mitem_uuid,       FALSE }
};

#ifdef ENABLE_CONSOLEHELPER
/* Applet actions menu rows */
static VmAppletActmenuRow vm_amenu_rows[] =
{
    { "Details...",         vm_amenu_details_cb,   VM_STOCK_DO_DETAILS,  1, VM_SMASK_ALL },
    { "Configuration...",   vm_amenu_edit_cb,      VM_STOCK_DO_EDIT,     1, VM_SMASK_ALL },
    { NULL, NULL, VM_STOCK_NONE, 0 }, /* separator */
    { "Suspend",            vm_amenu_suspend_cb,   VM_STOCK_DO_SUSPEND,  0, VM_SMASK_RUNNING|VM_SMASK_BLOCKED },
    { "Resume",             vm_amenu_resume_cb,    VM_STOCK_DO_RESUME,   0, VM_SMASK_PAUSED },
    { "Destroy",            vm_amenu_destroy_cb,   VM_STOCK_DO_DESTROY,  0, VM_SMASK_ALL },
    { "Shutdown",           vm_amenu_shutdown_cb,  VM_STOCK_DO_SHUTDOWN, 0, VM_SMASK_RUNNING|VM_SMASK_BLOCKED },
    { NULL, NULL, VM_STOCK_NONE, 0 }, /* separator */
    { "Terminal...",        vm_amenu_terminal_cb,  VM_STOCK_DO_TERMINAL, 0, VM_SMASK_RUNNING|VM_SMASK_BLOCKED },
    { "Virtual Console...", vm_amenu_console_cb,   VM_STOCK_DO_CONSOLE,  0, VM_SMASK_RUNNING|VM_SMASK_BLOCKED },
};
#endif

/* 
 * Prints and deallocates GError 
 */
static void
hmm_error(GError **error)
{
    _F();
	fprintf(stderr, "%s\n", (*error)->message);
	g_error_free (*error);
    *error = NULL;
}

/*
 * Initializes icons stock
 */
static void  
vm_init_icons(void)
{
    GtkIconFactory *icon_factory;
    gint i;

    _F();
    
	icon_factory = gtk_icon_factory_new();

	for (i = 0; i < G_N_ELEMENTS(vm_stock_icons); i++) 
    {
		VmAppletStockIcon icon = vm_stock_icons[i];
		GtkIconSource *icon_source = gtk_icon_source_new();
		GtkIconSet *icon_set = gtk_icon_set_new();

		gtk_icon_source_set_filename(icon_source, icon.filename);
		gtk_icon_set_add_source(icon_set, icon_source);
		gtk_icon_factory_add(icon_factory, icon.stock_id, icon_set);
		gtk_icon_source_free(icon_source);
		gtk_icon_set_unref(icon_set);
	}

	gtk_icon_factory_add_default(icon_factory);
	g_object_unref(G_OBJECT(icon_factory));
    
    gtk_window_set_default_icon_name("vm-applet");
}

/* 
 * Pops up an error dialog on the same screen as 'applet' saying 'msg'.
 */
static void
vm_error_dialog(GtkWidget *applet, char *msg)
{
    GtkWidget *dialog;
    
    _F();
    
    dialog = gtk_message_dialog_new( NULL, 0, GTK_MESSAGE_ERROR,
                                     GTK_BUTTONS_OK, msg);
    
    gtk_window_set_screen(GTK_WINDOW (dialog),
                           gtk_widget_get_screen (applet));
    
    gtk_window_set_title(GTK_WINDOW(dialog), "VM Applet");
    
    g_signal_connect_swapped( GTK_OBJECT (dialog), "response",
                              G_CALLBACK (gtk_widget_destroy),
                              GTK_OBJECT (dialog) );
    gtk_widget_show_all( dialog );
}

/* 
 * Common function invoked by the 'Help' context menu item and the 'Help'
 * button in the preferences dialog.
 */
void
vm_show_help( VmAppletCtl *ctl, const char *section )
{
    GError *error = NULL;
    
    _F();
    
    gnome_help_display_on_screen( "vm-applet", section,
                                gtk_widget_get_screen( ctl->applet ),
                                &error);
    if( error )
    {
        char *message;

        message = g_strdup_printf( _("There was an error displaying help: %s"),
                                 error->message );
        vm_error_dialog (ctl->applet, message);
        g_error_free( error );
        g_free( message );
    }
}

/* 
 * Called when the user selects the 'help' menu item.
 */
static void
vm_help_cb( BonoboUIComponent *uic, VmAppletCtl *ctl, const char *verb )
{
    _F();
    vm_show_help( ctl, NULL );
}

/* 
 * Called when the user selects the 'about' menu item.
 */
static void
vm_about_cb( BonoboUIComponent *uic, VmAppletCtl *ctl, const char *verb )
{
    GdkPixbuf *pixbuf = NULL;
    GError *error = NULL;
        
    const gchar *authors[] = {
        "Karel Zak <kzak@redhat.com>", 
        NULL
    };
    
    _F();
    
    pixbuf = gdk_pixbuf_new_from_file(VM_ICON, &error);
    if (error)
            hmm_error(&error);

    gtk_show_about_dialog( NULL,
        "name",                _("VM Applet"), 
        "version",             VERSION,
        "copyright",           "\xC2\xA9 2005, 2006 Red Hat Inc. ",
        "comments",            _("A utility for monitoring locally-running Virtual domains."),
        "authors",             authors,
        "translator-credits",  _("translator-credits"),
        "logo",               pixbuf,
        NULL );

    g_object_unref(pixbuf); /* XXX */
}

/* 
 * Called when the user selects the 'properties' menu item.
 */
static void
vm_prop_cb( BonoboUIComponent *uic, VmAppletCtl *ctl, const char *verb )
{
    _F();
    vm_prop_show(ctl);
}
   
/* 
 * Called when the user resizes the gnome panel
 */ 
static void
vm_size_cb( PanelApplet *applet, GtkAllocation *allocation, VmAppletCtl *ctl)
{
    _F();
    vm_icon_update(ctl, MIN(allocation->width, allocation->height), FALSE);    
}
    
static int
vm_connections_refresh(VmAppletCtl *ctl)
{
    VmAppletConn *conn = ctl->conn;
    
    _F();

    ctl->active_conns = 0;

    while(conn)
    {
        if (!conn->ptr)
        {
            if (conn->readonly)
                conn->ptr = virConnectOpenReadOnly(conn->uri);
            else
                conn->ptr = virConnectOpen(conn->uri);

            if (conn->ptr)
                virNodeGetInfo(conn->ptr, &conn->ninfo);
        }
        if (conn->ptr)
            ctl->active_conns++;
        conn = conn->next;
    }
    return ctl->active_conns ? TRUE : FALSE;
}

static void
vm_connections_destroy(VmAppletCtl *ctl)
{
    VmAppletConn *conn = ctl->conn;

    _F();

    while(conn)
    {
        if (conn->dom)
            vm_domainslist_free(conn->dom);
        if (conn->ptr)
            virConnectClose(conn->ptr);
        if (conn->uri)
            g_free(conn->uri);
        conn = conn->next;
    }
}

static VmAppletConn *
vm_connection_add(VmAppletCtl *ctl, char *uri, gboolean readonly)
{
    VmAppletConn *conn, *c = ctl->conn;

    _F();

    conn = g_new0(VmAppletConn, 1);
    if (!conn)
        return NULL;
    conn->uri = g_strdup(uri);
    conn->readonly = readonly;

    while(c && c->next)
        c = c->next;
    if (!c)
        ctl->conn = conn;
    else
        c->next = conn;

    return conn;
}

/*
 * Creates applet control struct
 */
static VmAppletCtl *
vm_ctl_new(PanelApplet *applet)
{
    GError *error = NULL;
    VmAppletCtl *ctl;
    int i;

    _F();
    
    /* allocate ctl */
    ctl = g_new0(VmAppletCtl, 1);
    ctl->applet = GTK_WIDGET(applet);
    ctl->uid = getuid();
    ctl->instance = instances;
    panel_applet_add_preferences(applet, "/schemas/apps/vm-applet/prefs", NULL);
    
    /* delay */
    ctl->delay = panel_applet_gconf_get_int(applet, VM_GCONF_DELAY, &error);
    if (error)
        hmm_error(&error);
    _D("read gconf: %s=%d", VM_GCONF_DELAY, ctl->delay);
    ctl->delay = ctl->delay <= 0 ? 1000 : ctl->delay;
    
    /* terminal cmd */
    ctl->terminal_cmd = panel_applet_gconf_get_string(applet, 
                 VM_GCONF_TERMINAL_CMD, &error);
    if (error)
        hmm_error(&error);
    _D("read gconf: %s=%s", VM_GCONF_TERMINAL_CMD, ctl->terminal_cmd);

    /* terminal enable */
    ctl->terminal_enb = panel_applet_gconf_get_bool(applet, 
                VM_GCONF_TERMINAL_ENABLE, &error);
    if (error)
        hmm_error(&error);
    _D("read gconf: %s=%s", 
                VM_GCONF_TERMINAL_ENABLE,
                ctl->terminal_enb ? "TRUE" : "FALSE");
 
    /* initialize menu description */
    for (i=0; i < G_N_ELEMENTS(vm_menu_desc); i++)
    {
        VmAppletMenuDesc *desc = &vm_menu_desc[i];
        gboolean val;
      
        desc->ctl = ctl;
        
        val = panel_applet_gconf_get_bool(applet, desc->gconf, &error);
        _D("read gconf: %s=%s", desc->gconf, val ? "TRUE" : "FALSE");

        if (error)
            hmm_error(&error);
        else
            desc->enable = val; 
    }
    
    ctl->desc = vm_menu_desc;
    ctl->ndescs = G_N_ELEMENTS(vm_menu_desc);
    
    /* TODO: use gconf and Preferences dialog */
    vm_connection_add(ctl, VM_DEFAULT_URI, ctl->uid ? TRUE : FALSE);

    vm_connections_refresh(ctl);
   
    return ctl;
}

/*
 * (Re)sets applet timer
 */
void
vm_timer_refresh(VmAppletCtl *ctl)
{
    _F();
    
	if (ctl->pixtimer) 
        g_source_remove(ctl->pixtimer);

    _D("setting delay to: %d (instance: %d)", ctl->delay, ctl->instance);
    ctl->pixtimer = g_timeout_add (ctl->delay, (GSourceFunc) vm_update, ctl);
}

/*
 * Removes (but not deallocates) domain from list of domains
 */
static VmAppletDomain *
vm_domainslist_rem(VmAppletDomain *list, VmAppletDomain *dom)
{
    VmAppletDomain *d = list;
    
    _F();
         
    if (!list)
        return NULL;
    
    if (dom==list)
        return list->next;
    
    while(d)
    {
        if (d->next==dom)
        {
            d->next = dom->next;
            dom->next = NULL;
            break;
        }
        d = d->next;
    }
    return list;
}

/*
 * Adds domain to the list of domains
 */
static VmAppletDomain *
vm_domainslist_add(VmAppletDomain *list, VmAppletDomain *dom)
{
    VmAppletDomain *d = list;
    
    _F();

    dom->next = NULL;
    
    if (!list)
        return dom;
    
    while(d)
    {
        if (!d->next)
        {
            d->next = dom;
            break;
        }
        d = d->next;
    }
    return list;
}

/*
 * Looks up domain in the list of domains
 */
static VmAppletDomain *
vm_domainslist_lookup(VmAppletDomain *list, char *uuid)
{
    _F();

    for (; list; list=list->next)
    {
        if (strcmp(list->uuid, uuid)==0)
            return list;
    }
    return NULL;
}

/*
 * Deallocates all domains in the list
 */
static void
vm_domainslist_free(VmAppletDomain *list)
{
    VmAppletDomain *d = list;

    _F();
    
    while(d)
    {
        VmAppletDomain *next = d->next;
        vm_domain_free(d);
        d = next;
    }
}
 
/*
 * Sets real state of domain
 * -- we assume the domain that uses CPU is not blocked. We cannot
 *    use only the domain->state field, because it's alwais "blocked"
 *    on sigle CPU boxes where is just running this code.
 */ 
static void
vm_domain_set_activity(VmAppletCtl *ctl, VmAppletDomain *dom)
{
    virDomainState old = dom->activity;
    
    _F();
    
    if (dom->state==VIR_DOMAIN_BLOCKED) 
    { 
        if ((ctl->delay * dom->cputime_ct) > VM_BLOCKED_TIME) 
            dom->activity = VIR_DOMAIN_BLOCKED; 
        else 
            dom->activity = VIR_DOMAIN_RUNNING; 
    } 
    else if (dom->state==VIR_DOMAIN_NOSTATE)
    {
        /* better old state or running than nothing */
        dom->activity = old==VIR_DOMAIN_NOSTATE ? VIR_DOMAIN_RUNNING : old;
    }
    else
        dom->activity = dom->state;             /* other than running or blocked state */

    _D("domain-activity: %s: '%s'", dom->name, vm_domainstate_to_string(dom->activity));

    if (old != dom->activity)
    {
        _D("domain-changed: %s: activity", dom->name);
        dom->changed |= VM_DATA_ACTIVITY;
    }
}

/*
 * Allocates new domain struct and load info about the domain
 */
static VmAppletDomain *
vm_domain_new(VmAppletCtl *ctl, virDomainPtr vir, char *uuid, VmAppletConn *conn)
{
    VmAppletDomain *dom;
    virDomainInfo info;
    char *p;
    
    _F();
    
    dom = g_new0(VmAppletDomain, 1);
    dom->ctl = ctl;
    dom->changed = 0;
    dom->conn = conn;
    
    dom->id = virDomainGetID(vir);
    dom->changed |= VM_DATA_ID;

    if (uuid)
        strcpy(dom->uuid, uuid);
    else
        virDomainGetUUIDString(vir, dom->uuid);
    dom->changed |= VM_DATA_UUID;
    
    dom->name = g_strdup(virDomainGetName(vir));
    dom->changed |= VM_DATA_NAME;
    
    _D("domain-new: %s: %s / %s", dom->conn->uri, dom->name, dom->uuid);
    
    if ((p = virDomainGetOSType(vir)))
    {
        dom->ostype = g_strdup(p);
        dom->changed |= VM_DATA_OSTYPE;
        free(p);
    }
    if (virDomainGetInfo(vir, &info)==0)
    {
        dom->state = info.state;
        dom->changed |= VM_DATA_STATE;
        
        dom->maxmem = info.maxMem;
        dom->changed |= VM_DATA_MAXMEM;
        
        dom->usemem = info.memory;
        dom->changed |= VM_DATA_USEDMEM;
        
        dom->ncpus = info.nrVirtCpu;
        dom->changed |= VM_DATA_NCPUS;
        
        dom->cputime = info.cpuTime;
        dom->changed |= VM_DATA_CPUTIME;
    }
    
    vm_domain_set_activity(ctl, dom);
    return dom;
}

/*
 * Refreshs information about the domain
 */
static void
vm_domain_refresh(VmAppletCtl *ctl, VmAppletDomain *dom, virDomainPtr vir, char *uuid)
{
    virDomainInfo info;
    const char *p;
    char xuuid[VM_UUID_SIZ];
    
    _F();
    
    _D("domain-refresh: %s : %s", dom->conn->uri, dom->name);
    
    dom->changed = 0;
    
    if (dom->id!=virDomainGetID(vir))
    {
        _D("domain-changed: %s: id", dom->name);
        dom->id = virDomainGetID(vir);
        dom->changed |= VM_DATA_ID;
    }
    if (!uuid)
    {
        if (virDomainGetUUIDString(vir, xuuid)==0)
            uuid = xuuid;
    }
    if (uuid && strcmp(dom->uuid, uuid))
    {
        _D("domain-changed: %s: uuid", dom->name);
        dom->changed |= VM_DATA_UUID;
        strcpy(dom->uuid, uuid);
    }
    p = virDomainGetName(vir);
    if (p && strcmp(dom->name, p))
    {
        _D("domain-changed: %s: name", dom->name);
        g_free(dom->name);
        dom->name = g_strdup(p);
        dom->changed |= VM_DATA_NAME;
    }
    p = virDomainGetOSType(vir);
    if (p && strcmp(dom->ostype, p))
    {
        _D("domain-changed: %s: ostype", dom->name);
        g_free(dom->ostype);
        dom->ostype = g_strdup(p);
        dom->changed |= VM_DATA_OSTYPE;
    }
    if (p)
        free((void *)p);
    
    if (virDomainGetInfo(vir, &info)==0) 
    {
        if (dom->state != info.state)
        {
            _D("domain-changed: %s: state", dom->name);
            dom->state = info.state;
            dom->changed |= VM_DATA_STATE;
        }
        if (dom->maxmem != info.maxMem)
        {
            _D("domain-changed: %s: maxmem", dom->name);
            dom->maxmem = info.maxMem;
            dom->changed |= VM_DATA_MAXMEM;
        }
        if (dom->usemem != info.memory)
        {
            _D("domain-changed: %s: usemem", dom->name);
            dom->usemem = info.memory;
            dom->changed |= VM_DATA_USEDMEM;
        }
        if (dom->ncpus != info.nrVirtCpu)
        {
            _D("domain-changed: %s: ncpus", dom->name);
            dom->ncpus = info.nrVirtCpu;
            dom->changed |= VM_DATA_NCPUS;
        }
        if (dom->cputime != info.cpuTime)
        {
            _D("domain-changed: %s: cputime", dom->name);
            dom->cputime = info.cpuTime;
            dom->changed |= VM_DATA_CPUTIME;
            dom->cputime_ct = 0;
        }
        else
            dom->cputime_ct++; 
    }
    vm_domain_set_activity(ctl, dom);
}

/*
 * Updates all domains 
 */
static void
vm_domains_update(VmAppletCtl *ctl)
{
    VmAppletConn *conn = ctl->conn;

    _F();

    ctl->ndoms = 0;

    while(conn)
    {
        ctl->ndoms += vm_domains_update_from_conn(ctl, conn);
        conn = conn->next;
    }
}

static int
vm_domains_update_from_conn(VmAppletCtl *ctl, VmAppletConn *conn)
{
    VmAppletDomain *old, *new, *dom;
    int *ids, i;
    int ct = 0;

    _F();
    
    old = conn->dom;
    new = NULL;
    
    conn->ndoms = virConnectNumOfDomains(conn->ptr);
    if (conn->ndoms <= 0)
    {
       vm_domainslist_free(conn->dom);
       conn->dom = NULL;
       return 0;
    }
    ids = g_malloc(sizeof(int) * conn->ndoms);
    virConnectListDomains(conn->ptr, &ids[0], conn->ndoms);

    for(i=0; i < conn->ndoms; i++) 
    {
        virDomainPtr vir = virDomainLookupByID(conn->ptr, ids[i]);
        char uuid[VM_UUID_SIZ];
       
        if (virDomainGetUUIDString(vir, uuid) < 0)
            continue;
        
        if ((dom = vm_domainslist_lookup(old, uuid)))
        {
            old = vm_domainslist_rem(old, dom);
            vm_domain_refresh(ctl, dom, vir, uuid);
        }
        else
            dom =  vm_domain_new(ctl, vir, uuid, conn);
        
        new = vm_domainslist_add(new, dom);
        virDomainFree(vir);
        ct++;
    }

    if (old)
        vm_domainslist_free(old);
    conn->dom = new;
    g_free(ids);

    return ct;
}

/*
 * Deallocate the domain
 */
static void
vm_domain_free(VmAppletDomain *dom)
{
    _F();
    
    if (!dom)
        return;
    
    if (dom->ctl->popmenu && dom->menuitem) 
        gtk_widget_destroy(dom->menuitem);
    if (dom->name)
        g_free(dom->name);
    if (dom->ostype)
        g_free(dom->ostype);

    g_free(dom);
}

/*
 * Deallocates the applet control struct
 */
static void
vm_ctl_destroy(VmAppletCtl *ctl)
{
    _F();
    
    if (ctl->conn)
        vm_connections_destroy(ctl);

    g_free(ctl);
}

/*
 * Updates the tool tip on the gome-panel
 */
static void
vm_tooltips_update(VmAppletCtl *ctl)
{
	char *tooltip;
	
    _F();
    
    if (ctl->ndoms > 0)
    {
        if (ctl->ndoms > 1)
        	tooltip = g_strdup_printf (_("%d virtual domains"), ctl->ndoms);
        else
            tooltip = g_strdup_printf (_("%d virtual domain"), ctl->ndoms);
        gtk_tooltips_set_tip (ctl->applet_tip, ctl->applet, 
                tooltip, NULL);
        g_free (tooltip);
    }
    else if (!ctl->active_conns)
        gtk_tooltips_set_tip (ctl->applet_tip, ctl->applet, 
                _("Failed to connect to the hypervisor."), NULL);
    else
        gtk_tooltips_set_tip (ctl->applet_tip, ctl->applet, 
                _("Failed to load information about virtual domains."), NULL);
}

/* 
 * Creates or updates the applet 
 */
static gint
vm_update(VmAppletCtl *ctl)
{
    int ndoms = 0;

    _D("\n\n---- loop (pid: %d, instance: %d, delay: %d) ----", getpid(), ctl->instance, ctl->delay);
    _F();
    
    ndoms = ctl->ndoms;
    
    if (!ctl->active_conns)
        vm_connections_refresh(ctl);
    
    vm_domains_update(ctl);
    
    _D("ndoms=%d, ctl->ndoms=%d", ndoms, ctl->ndoms);
    
    if (ctl->image && ((ndoms<=0 && ctl->ndoms > 0) || (ndoms > 0 && ctl->ndoms<=0)))
        vm_icon_update(ctl, ctl->image_size, TRUE);
    
    if (ctl->popmenu && ctl->popmenu_up==1)
        vm_popmenu_update(ctl);

    vm_tooltips_update(ctl);
    
    _D("\n\n");
    
    return TRUE;
}

/*
 * Deallocates the applet (the end! ;-)
 */
static void
vm_destroy_cb(GtkWidget *widget, VmAppletCtl *ctl)
{   
    _F(); 
    
    g_source_remove(ctl->pixtimer);
    
    if (ctl->popmenu)
    {
        gtk_widget_destroy(ctl->popmenu);
        ctl->popmenu = NULL;
    }
    vm_ctl_destroy(ctl);

    --instances;
}


/* 
 * Counts pop menu position
 * -- the following function is adapted from gtkmenuitem.c 
 */
static void
vm_position_popmenu (GtkMenu *menu, gint *x, gint *y,
	       gboolean *push_in, gpointer ctl)
{
    GtkWidget *widget = ((VmAppletCtl *) ctl)->applet;
    GdkScreen *screen;
    gint twidth, theight, tx, ty;
    GtkTextDirection direction;
    GdkRectangle monitor;
    gint monitor_num;

    _F();
    
    g_return_if_fail (menu != NULL);
    g_return_if_fail (x != NULL);
    g_return_if_fail (y != NULL);

    if (push_in) *push_in = FALSE;

    direction = gtk_widget_get_direction (widget);

    twidth = GTK_WIDGET (menu)->requisition.width;
    theight = GTK_WIDGET (menu)->requisition.height;

    screen = gtk_widget_get_screen (GTK_WIDGET (menu));
    monitor_num = gdk_screen_get_monitor_at_window (screen, widget->window);
    if (monitor_num < 0) monitor_num = 0;
    gdk_screen_get_monitor_geometry (screen, monitor_num, &monitor);

    if (!gdk_window_get_origin (widget->window, &tx, &ty)) 
    {
	    g_warning ("Menu not on screen");
	    return;
    }

    tx += widget->allocation.x;
    ty += widget->allocation.y;

    if (direction == GTK_TEXT_DIR_RTL)
	tx += widget->allocation.width - twidth;

    if ((ty + widget->allocation.height + theight) <= monitor.y + monitor.height)
	ty += widget->allocation.height;
    else if ((ty - theight) >= monitor.y)
	ty -= theight;
    else if (monitor.y + monitor.height - (ty + widget->allocation.height) > ty)
	ty += widget->allocation.height;
    else
	ty -= theight;

    *x = CLAMP (tx, monitor.x, MAX (monitor.x, monitor.x + monitor.width - twidth));
    *y = ty;
    gtk_menu_set_monitor (menu, monitor_num);
}

#ifdef ENABLE_CONSOLEHELPER

static void
vm_amenu_action_exec(VmAppletDomain *domain, char *action, char *extra)
{
    pid_t pid; 
    
    _F();

    if (domain==NULL || action==NULL || *action=='\0')
        return;
    
    _D("exec-> %s on %s\n", action, domain->uuid);

    switch((pid=fork()))
    {
        case -1:    /* error */
            return;
        case 0:     /* child */
            if (extra)
                execlp(VM_WRAPPER_NAME, VM_WRAPPER_NAME, action, domain->uuid, extra, NULL);
            else
                execlp(VM_WRAPPER_NAME, VM_WRAPPER_NAME, action, domain->uuid, NULL);
            exit(EXIT_FAILURE);
        default: /* parent */
            waitpid(pid, NULL, 0);
    }
}

static void
vm_amenu_suspend_cb(GtkMenuItem *menuitem, gpointer data)
{
    vm_amenu_action_exec(((VmAppletCtl *) data)->actdom, "suspend", NULL);
}

static void
vm_amenu_resume_cb(GtkMenuItem *menuitem, gpointer data)
{
    vm_amenu_action_exec(((VmAppletCtl *) data)->actdom, "resume", NULL);
}

static void
vm_amenu_destroy_cb(GtkMenuItem *menuitem, gpointer data)
{
    vm_amenu_action_exec(((VmAppletCtl *) data)->actdom, "destroy", NULL);
}

static void
vm_amenu_shutdown_cb(GtkMenuItem *menuitem, gpointer data)
{
    vm_amenu_action_exec(((VmAppletCtl *) data)->actdom, "shutdown", NULL);
}

static void
vm_amenu_terminal_cb(GtkMenuItem *menuitem, gpointer data)
{
    VmAppletCtl *ctl = (VmAppletCtl *) data;
    GError *err = NULL;
    char **argv = NULL;
    char *cmd = NULL;
    pid_t pid;
    
    _F();

    /* TODO: ignore for remote domains */
    if (ctl->terminal_enb==FALSE || ctl->terminal_cmd==NULL || *ctl->terminal_cmd=='\0')
        return;
    
    if (!(cmd = g_strdup_printf(ctl->terminal_cmd, ctl->actdom->uuid)))
        goto End;

    if (!g_shell_parse_argv(cmd, NULL, &argv, &err))
    {
        hmm_error(&err);
        goto End;
    }
    
    switch((pid=fork()))
    {
        case -1:    /* error */
            return;
        case 0:     /* child */
            execvp(argv[0], argv);
            exit(EXIT_FAILURE);
        default: /* parent */
            waitpid(pid, NULL, 0);
    }

End:
    _D("done");
    if (argv)
        g_strfreev(argv);
    if (cmd)
        g_free(cmd);
}

/* deactivate also domains menu */
static void
vm_amenu_deactivate_cb(GtkMenuShell *menushell, VmAppletCtl *ctl)
{
    _F();
    gtk_menu_popdown(GTK_MENU(ctl->popmenu));
}

static void
vm_amenu_details_cb(GtkMenuItem *menuitem, gpointer data)
{
    VmAppletCtl *ctl = (VmAppletCtl *) data;

    _F();

    vm_mngr_method_call(ctl, "show_domain_performance", ctl->actdom, NULL);
}

static void
vm_amenu_edit_cb(GtkMenuItem *menuitem, gpointer data)
{
    VmAppletCtl *ctl = (VmAppletCtl *) data;

    _F();

    vm_mngr_method_call((VmAppletCtl *)data, "show_domain_editor", ctl->actdom, NULL);
}

static void
vm_amenu_console_cb(GtkMenuItem *menuitem, gpointer data)
{
    VmAppletCtl *ctl = (VmAppletCtl *) data;

    _F();

    vm_mngr_method_call(ctl, "show_domain_console", ctl->actdom, NULL);
}


static GtkWidget *
vm_amenu_new(gpointer data)
{
    GtkWidget *menu;
    int i;

    _F();

    menu = gtk_menu_new();
   
    for (i=0; i < G_N_ELEMENTS(vm_amenu_rows); i++)
    {
        VmAppletActmenuRow *row = &vm_amenu_rows[i];
        GtkWidget *item; 
            
        if (row->label)
        {
            item = gtk_image_menu_item_new_with_label(row->label); 
        
            if (row->stock_id)
            {
                GtkWidget *image = gtk_image_new_from_stock (row->stock_id, GTK_ICON_SIZE_MENU);
                gtk_image_menu_item_set_image(GTK_IMAGE_MENU_ITEM(item), image);
            } 
            gtk_widget_set_sensitive(item, TRUE);
            g_signal_connect(G_OBJECT(item), "activate", G_CALLBACK(row->callback), data);
        }
        else
            item = gtk_separator_menu_item_new();
        
        g_object_set_data(G_OBJECT(item), "vmrow", (gpointer) row);
        gtk_container_add(GTK_CONTAINER(menu), item);
    }    
    
	g_signal_connect(G_OBJECT(menu), "deactivate",
			G_CALLBACK(vm_amenu_deactivate_cb), data);

    gtk_widget_show_all(menu);
    return menu;
}

/*
 * The item sensitivity depends on the selected domain state. 
 * We don't offer all actions for all domains. The setting is 
 * done by VmAppletActmenuRow->mask (with exception for Xen 
 * dom0).
 */
static void
vm_amenu_item_sensitive(GtkWidget *item, gpointer data)
{
    VmAppletDomain *domain = (VmAppletDomain *) data;
    VmAppletActmenuRow *row;
     
    _F();

    row = (VmAppletActmenuRow *) g_object_get_data(G_OBJECT(item), "vmrow");

    if (!row->label)
        return; /* menu separator */
    
    if (domain->id==0 && row->dom0==0)
        /* dom0 */
        gtk_widget_set_sensitive(item, FALSE);
    /* TODO: gnore for remote domains */
    else if (row->callback==vm_amenu_terminal_cb && domain->ctl->terminal_enb==FALSE)
        /* special: terminal */
        gtk_widget_set_sensitive(item, FALSE);
    else if (!vm_state_is_valid(domain->activity))
        /* invalid state */
        gtk_widget_set_sensitive(item, FALSE);
    else if (row->mask & vm_state2mask(domain->activity))
        /* valid state and should be active by mask */
        gtk_widget_set_sensitive(item, TRUE);
    else
        /* valid state and should be inactive by mask */
        gtk_widget_set_sensitive(item, FALSE);
}

static gboolean
vm_menuitem_cb(GtkWidget *menuitem, GdkEventButton *event, gpointer data)

{
    VmAppletDomain *domain = (VmAppletDomain *) data;

    _F();

    if (domain==NULL || domain->ctl==NULL)
        return TRUE;
    
    if (!domain->ctl->actmenu)
        domain->ctl->actmenu = vm_amenu_new(domain->ctl);
    
    domain->ctl->actdom = domain;
    
    /* disable some items fro dom0 */
    gtk_container_foreach(GTK_CONTAINER(domain->ctl->actmenu), 
            vm_amenu_item_sensitive, domain);
    
	gtk_menu_popup (GTK_MENU (domain->ctl->actmenu), 
            NULL, NULL, NULL, NULL, event->button, event->time);

    return TRUE;
}

#endif /* ENABLE_CONSOLEHELPER */

static void
vm_dbus_method_error(VmAppletCtl *ctl, GError *error)
{
    char *message;
        
    _F();

    if (error->domain == DBUS_GERROR && error->code == DBUS_GERROR_REMOTE_EXCEPTION)
        message = g_strdup_printf(
                "Caught remote method exception %s: %s",
                dbus_g_error_get_name (error),
                error->message);
    else
        message = g_strdup_printf("D-BUS Error: %s\n", error->message);
    
    _D("D-BUS Error: %s", message);

    vm_error_dialog(ctl->applet, message);
    g_error_free( error );
    g_free( message );
}

static DBusGConnection *
vm_dbus_get_connection(VmAppletCtl *ctl, DBusBusType type)
{
    GError *error = NULL;
    DBusGConnection *connection;

    _F();

    connection = dbus_g_bus_get (type, &error);
    
    if (connection == NULL)
    {
        char *message = g_strdup_printf(
                            _("Failed to open connection to D-BUS: %s"),
                            error->message );
        vm_error_dialog(ctl->applet, message);
        g_error_free( error );
        g_free( message );
    }
    return connection;
}

static gboolean
vm_mngr_run_cb(GtkWidget *menuitem, GdkEventButton *event, gpointer data)

{
    _F();
    vm_mngr_method_call((VmAppletCtl *) data, "show_connect", NULL, NULL);
    return TRUE;
}

static gboolean
vm_mngr_method_call(VmAppletCtl *ctl, const char *methodname, 
                VmAppletDomain *dom, const char *uri)
{
    DBusGConnection *connection;
    DBusGProxy *proxy;
    GError *error = NULL;
    gboolean ret;

    _F();

    if (!(connection = vm_dbus_get_connection(ctl, DBUS_BUS_SESSION)))
        return TRUE;
    
    proxy = dbus_g_proxy_new_for_name (connection,
                    VM_DBUS_VIRTMANAGER_SERVICE,
                    VM_DBUS_VIRTMANAGER_PATH,
                    VM_DBUS_VIRTMANAGER_INTERFACE);

    _D("D-BUS proxy: %s", proxy ? "yes" : "no");

    if (!proxy)
        return FALSE;

    if (dom)
        ret = dbus_g_proxy_call(proxy, methodname, &error,
            G_TYPE_STRING, uri ? uri : dom->conn->uri, 
            G_TYPE_STRING, dom->uuid,
            G_TYPE_INVALID,
            G_TYPE_INVALID);
    else if (uri)
        ret = dbus_g_proxy_call(proxy, methodname, &error,
            G_TYPE_STRING, uri,
            G_TYPE_INVALID,
            G_TYPE_INVALID);
    else
        ret = dbus_g_proxy_call(proxy, methodname, &error,
            G_TYPE_INVALID,
            G_TYPE_INVALID);

    if (!ret)
    {
            vm_dbus_method_error(ctl, error);
            g_object_unref (proxy);
            return FALSE;
    }
    g_object_unref (proxy);
    return TRUE;
}

static gboolean
vm_mngr_new_cb(GtkWidget *menuitem, GdkEventButton *event, gpointer data)
{
    VmAppletCtl *ctl = (VmAppletCtl *) data;

    _F();

    /* XXX: ark for URI */
    vm_mngr_method_call(ctl, "show_domain_creator", NULL, VM_DEFAULT_URI);
    return TRUE;
}

/*
 * Blocked or really dead?
 */
#define VM_IS_BLOCKED(_ctl, _dom) \
            ((_ctl)->delay * (_dom)->cputime_ct > VM_BLOCKED_TIME)

/*
 * Converts the domain state to string
 */
static char *
vm_domainstate_to_string(virDomainState state) 
{
    _F();
    switch (state) {
        case VIR_DOMAIN_BLOCKED:
            return _("blocked");
        case VIR_DOMAIN_RUNNING:
            return _("running");
        case VIR_DOMAIN_PAUSED:
            return _("paused");
        case VIR_DOMAIN_SHUTDOWN:
            return _("in shutdown");
        case VIR_DOMAIN_SHUTOFF:
            return _("shut off");
        default:
            return "unknown";    /* = dom0 state */
    }
    return NULL;
}

/*
 * Converts the domain state to icon stock Id
 */
static char *
vm_domainstate_to_stockid(virDomainState state) 
{
    _F();
    switch (state) {
        case VIR_DOMAIN_BLOCKED:
            return VM_STOCK_BLOCKED;
        case VIR_DOMAIN_RUNNING:
            return VM_STOCK_RUNNING;
        case VIR_DOMAIN_PAUSED:
            return VM_STOCK_PAUSED;
        case VIR_DOMAIN_SHUTDOWN:
            return VM_STOCK_SHUTDOWN;
        case VIR_DOMAIN_SHUTOFF:
            return VM_STOCK_SHUTOFF;
        default:
            return VM_STOCK_UNKNOWN;    /* = dom0 state */
    }
    return NULL;
}

static GtkWidget *
vm_mitem_os(VmAppletDomain *domain, VmAppletMenuDesc *desc, GtkWidget *item)
{
    _F();

    if (item)
        gtk_label_set_text(GTK_LABEL(item), domain->ostype);
    else
    {
        item = gtk_label_new(domain->ostype);
        gtk_misc_set_alignment (GTK_MISC(item), 0.0, 0.5);
    }
    return item;
}


static GtkWidget *
vm_mitem_state(VmAppletDomain *domain, VmAppletMenuDesc *desc, GtkWidget *item)
{
    _F();

    if (item)
        gtk_label_set_text(GTK_LABEL(item), vm_domainstate_to_string(domain->state));
    else
    {
        item = gtk_label_new(vm_domainstate_to_string(domain->state));
        gtk_misc_set_alignment (GTK_MISC(item), 0.0, 0.5);
    }
    return item;
}

static GtkWidget *
vm_mitem_activity(VmAppletDomain *domain, VmAppletMenuDesc *desc, GtkWidget *item)
{
    char *id;
    _F();

    if (!(id = vm_domainstate_to_stockid(domain->activity)))
    {
        if (domain->id==0)      /* dom0 */
            id = vm_domainstate_to_stockid(VIR_DOMAIN_RUNNING);
        else
            id = vm_domainstate_to_stockid(0);
    }
   
    if (item)
        gtk_image_set_from_stock(GTK_IMAGE(item), id, GTK_ICON_SIZE_MENU); 
    else
    {
        item =  gtk_image_new_from_stock (id, GTK_ICON_SIZE_MENU);
        gtk_misc_set_alignment (GTK_MISC(item), 0.5, 0.5);
    }
    return item;
}

static GtkWidget *
vm_mitem_id(VmAppletDomain *domain, VmAppletMenuDesc *desc, GtkWidget *item)
{
    char *str;
    _F();

    str = g_strdup_printf("%d", domain->id);
    if (item)
        gtk_label_set_text(GTK_LABEL(item), str);
    else
    {
        item = gtk_label_new(str);
        gtk_misc_set_alignment (GTK_MISC(item), 1.0, 0.5);
    }
    g_free(str);
    return item;
}

static GtkWidget *
vm_mitem_name(VmAppletDomain *domain, VmAppletMenuDesc *desc, GtkWidget *item)
{
    _F();

    if (item)
        gtk_label_set_text(GTK_LABEL(item), domain->name);
    else
    {
        item = gtk_label_new(domain->name);
        gtk_misc_set_alignment (GTK_MISC(item), 0.0, 0.5);
    }
    return item;
}

static GtkWidget *
vm_mitem_uuid(VmAppletDomain *domain, VmAppletMenuDesc *desc, GtkWidget *item)
{
    _F();

    if (item)
        gtk_label_set_text(GTK_LABEL(item), domain->uuid);
    else
    {
        item = gtk_label_new(domain->uuid);
        gtk_misc_set_alignment (GTK_MISC(item), 0.0, 0.5);
    }
    return item;
}

static GtkWidget *
vm_mitem_maxmem(VmAppletDomain *domain, VmAppletMenuDesc *desc, GtkWidget *item)
{
    char *str;
    _F();

    str = g_strdup_printf("%lu MB", (unsigned long)(domain->maxmem / 1024UL));
    if (item)
        gtk_label_set_text(GTK_LABEL(item), str);
    {
        item = gtk_label_new(str);
        gtk_misc_set_alignment (GTK_MISC(item), 1.0, 0.5);
    }
    g_free(str);
    return item;
}

static GtkWidget *
vm_mitem_usemem(VmAppletDomain *domain, VmAppletMenuDesc *desc, GtkWidget *item)
{
    char *str;
    _F();

    str = g_strdup_printf("%lu MB (%2.2f%%)", 
                (unsigned long)(domain->usemem / 1024UL),
                domain->usemem * 100 / (double) domain->conn->ninfo.memory);
    if (item)
        gtk_label_set_text(GTK_LABEL(item), str);
    else
    {
        item = gtk_label_new(str);
        gtk_misc_set_alignment (GTK_MISC(item), 1.0, 0.5);
    }
    g_free(str);
    return item;
}

static GtkWidget *
vm_mitem_ncpus(VmAppletDomain *domain, VmAppletMenuDesc *desc, GtkWidget *item)
{
    char *str;
    _F();

    str = g_strdup_printf("%d %s", domain->ncpus, domain->ncpus > 0 ? "CPUs" : "CPU");
    if (item)
        gtk_label_set_text(GTK_LABEL(item), str); 
    else
    {
        item = gtk_label_new(str);
        gtk_misc_set_alignment (GTK_MISC(item), 1.0, 0.5);
    }
    g_free(str);
    return item;
}

static GtkWidget *
vm_mitem_cputime(VmAppletDomain *domain, VmAppletMenuDesc *desc, GtkWidget *item)
{
    char *str = NULL;
    float cputime = domain->cputime;

    _F();

    if (cputime != 0) 
    {
        cputime /= 1000000000;
        str = g_strdup_printf("%.1fs", cputime);
    }
    if (item)
        gtk_label_set_text(GTK_LABEL(item), str); 
    else
    {
        item = gtk_label_new(str);
        gtk_misc_set_alignment (GTK_MISC(item), 1.0, 0.5);
    }
    g_free(str);
    return item;
}

/*
 * Creates menu item
 */
static GtkWidget *
vm_popmenuitem_new(VmAppletCtl *ctl, VmAppletDomain *dom, int pos)
{
    int i;
    GtkWidget *box;
    GtkWidget *menuitem;

    _F();
    
    menuitem = gtk_menu_item_new ();
    gtk_widget_set_sensitive (menuitem, TRUE);
    gtk_menu_shell_insert(GTK_MENU_SHELL(ctl->popmenu), menuitem, pos);

#ifdef ENABLE_CONSOLEHELPER
	g_signal_connect(G_OBJECT(menuitem), "button-press-event", G_CALLBACK(vm_menuitem_cb), dom);
#endif
    
    box = gtk_hbox_new(FALSE, 12);
    gtk_container_add (GTK_CONTAINER(menuitem), box);
            
    for(i=0; i < ctl->ndescs; i++)
    {
        VmAppletMenuDesc *desc = &ctl->desc[i];  
        GtkWidget *subitem;
        
        if (!desc->enable)
            continue;
        
        subitem = desc->builder(dom, desc, NULL);
        if (!subitem)
        {
            _D("popmenu-item-new: %s: no item: %s", dom->name, desc->name);
            subitem = gtk_label_new("");    /* dummy column */
        }
        gtk_box_pack_start(GTK_BOX(box), subitem, FALSE, FALSE, 0);
        gtk_size_group_add_widget(desc->sizegroup, subitem);
    }
    gtk_widget_show_all (menuitem);
    
    return menuitem;
}

/*
 * Updates fields in the menu item
 */
static GtkWidget *
vm_popmenuitem_update(VmAppletCtl *ctl, VmAppletDomain *dom, GtkWidget *menuitem)
{
    GtkWidget *box;
    GList *list, *li;
    int i=0;

    
    _F();
    
    if (!dom->changed)
        return menuitem;
    
    /* get Box */
    if (!(list = gtk_container_get_children(GTK_CONTAINER(menuitem))))
        return menuitem;
    li = g_list_first(list);
    box = li->data;
    g_list_free(list);
    
    /* get Items */
    list = gtk_container_get_children(GTK_CONTAINER(box));
    _D("popmenu-item-update: %s: found: %d subitems", dom->name, list ? g_list_length(list) : 0);
    
    li = g_list_first(list);
    for(i=0; i < ctl->ndescs; i++)
    {
        VmAppletMenuDesc *desc = &ctl->desc[i];  
        GtkWidget *subitem;
                
        if (!desc->enable)
        {
            _D("popmenu-item-disabled: %s: %s", dom->name, desc->name);
            continue;
        }
        if (li==NULL)
        {
            _D("*** out of list! ***");
            break;
        }
        subitem = li->data;
        li = g_list_next(li);
        
        if (!(dom->changed & desc->did))
        {
            _D("popmenu-item-nochange: %s: %s", dom->name, desc->name);
            continue;
        }
        _D("popmenu-item-change: %s: %s", dom->name, desc->name);
        desc->builder(dom, desc, subitem);
    }
    return menuitem;
}

/*
 * Updates all items in the menu
 */
static void
vm_popmenu_update(VmAppletCtl *ctl)
{
    VmAppletConn *conn = ctl->conn;
    int pos = 0;
    
    _F();
    
    while(conn)
    {
        VmAppletDomain *dom;

        for(dom=conn->dom; dom; dom=dom->next, pos++)
        {
            if (dom->menuitem==NULL)
            {
                dom->menuitem = vm_popmenuitem_new(ctl, dom, pos);
            }
            else if (dom->changed==0)
            {
                _D("popmenu-nochange: %s", dom->name);
                continue;
            }
            else 
            {
                _D("popmenu-change: %s", dom->name);
                vm_popmenuitem_update(ctl, dom, dom->menuitem);
            }
            dom->changed = 0;
        }
        conn = conn->next;
    }
}

void
vm_popmenu_destroy_items(VmAppletCtl *ctl)
{
    VmAppletConn *conn = ctl->conn;
   
    _F();
    
    while(conn)
    {
        VmAppletDomain *dom;

        for(dom=conn->dom; dom; dom=dom->next)
        {
            if (dom->menuitem)
                gtk_widget_destroy(dom->menuitem);
            dom->menuitem = NULL;
        }
        conn = conn->next;
    }
}

/*
 * Called when the menu pop down
 */
static void
vm_popmenu_deactivate_cb(GtkMenuShell *menushell, VmAppletCtl *ctl)
{
    _F();
    ctl->popmenu_up = 0;
}

/*
 * Don't use native gtk_image_menu_item_new_with_label(), it allocates 
 * extra space for menuitem icon.
 */
static GtkWidget *
vm_gtk_image_menu_item_new_with_label(const char *label, const char *stock_id)
{
    GtkWidget *menuitem;
    GtkWidget *subitem;
    GtkWidget *box;
    
    _F();
    
    menuitem = gtk_menu_item_new();
    box = gtk_hbox_new(FALSE, 12);
    
    if ((subitem = gtk_image_new_from_stock(stock_id, GTK_ICON_SIZE_MENU)))
    {
        gtk_misc_set_alignment (GTK_MISC(subitem), 0.5, 0.5);
        gtk_box_pack_start(GTK_BOX(box), subitem, FALSE, FALSE, 0);
    }
    subitem = gtk_label_new(label);
    gtk_box_pack_start(GTK_BOX(box), subitem, FALSE, FALSE, 0);
    
    gtk_container_add (GTK_CONTAINER(menuitem), box);
    gtk_widget_set_sensitive (menuitem, TRUE);

    return menuitem;
}


/*
 * Creates the applet pop menu
 */ 
static void
vm_popmenu_new(VmAppletCtl *ctl)
{
    GtkWidget *menuitem;
    int i;
   
    _F();
       
    if (ctl->popmenu)
        return;
    
    ctl->popmenu = gtk_menu_new ();
    ctl->popmenu_up = 0;

    for(i=0; i < ctl->ndescs; i++)
        ctl->desc[i].sizegroup = gtk_size_group_new(GTK_SIZE_GROUP_HORIZONTAL);

    /* separator */
    menuitem = gtk_separator_menu_item_new();
    gtk_widget_show(menuitem);
    gtk_container_add(GTK_CONTAINER(ctl->popmenu), menuitem);
    
    /* VM Manager: run */
    menuitem = vm_gtk_image_menu_item_new_with_label(
                    "Run Virtual Machine Manager...", GTK_STOCK_EXECUTE);
    gtk_widget_show_all(menuitem);
    gtk_container_add(GTK_CONTAINER(ctl->popmenu), menuitem);
    g_signal_connect(G_OBJECT(menuitem), "button-press-event", 
                    G_CALLBACK(vm_mngr_run_cb), ctl);

    /* VM Manager: new domain */
    menuitem = vm_gtk_image_menu_item_new_with_label(
                    "Create A New Virtual Machine...", GTK_STOCK_NEW);
    gtk_widget_show_all(menuitem);
    gtk_container_add(GTK_CONTAINER(ctl->popmenu), menuitem);
    g_signal_connect(G_OBJECT(menuitem), "button-press-event", 
                    G_CALLBACK(vm_mngr_new_cb), ctl);
  
	g_signal_connect(G_OBJECT(ctl->popmenu), "deactivate",
			G_CALLBACK(vm_popmenu_deactivate_cb), ctl);

    vm_popmenu_update(ctl);
}

/*
 * Called when user press the button on the gnome applet
 */
static gboolean
vm_button_cb(   GtkWidget *widget, 
                        GdkEventButton *event, 
                        VmAppletCtl *ctl)
{
    _F();

    if (event->button != 1)
        return FALSE;
    
    if (ctl->ndoms > 0)
    {    
        if (!ctl->popmenu)
            vm_popmenu_new(ctl);
        else 
            vm_popmenu_update(ctl);
        
        ctl->popmenu_up = 1;
        
        gtk_menu_popup (GTK_MENU (ctl->popmenu), NULL, NULL,
                    vm_position_popmenu, ctl,
                    event->button, event->time);
    }    
    return TRUE;
}

/*
 * Updates the incon on the gnome panel
 * -- called when user resizes the panel
 */
static void
vm_icon_update(VmAppletCtl *ctl, int sz, gboolean refresh)
{
    GdkPixbuf *pixbuf;
    
    _F();
    
    if (!refresh)
    {
        if (sz==ctl->image_size)
            return;
    }
    ctl->image_size = sz;
    
    if (refresh && ctl->image)
    {
         gtk_widget_destroy(ctl->image);
         ctl->image = NULL;
    }
    if (!ctl->image)
    {
        GError *error = NULL;
        
        _D("loading new icon");
        ctl->pixbuf = gdk_pixbuf_new_from_file(ctl->ndoms > 0 ? VM_ICON : VM_XICON, &error);
        if (error)
            hmm_error(&error);
       
        ctl->image = gtk_image_new();
        gtk_container_add(GTK_CONTAINER(ctl->applet), ctl->image);
    }
    
    pixbuf = gdk_pixbuf_scale_simple(ctl->pixbuf, sz, sz, GDK_INTERP_BILINEAR);
    gtk_image_set_from_pixbuf(GTK_IMAGE(ctl->image), pixbuf);

    g_object_unref(pixbuf);

    gtk_widget_show(ctl->image);
}

/* 
 * Called by the factory to fill in the fields for the applet.
 */
static gboolean
vm_setup (PanelApplet *applet)
{
    VmAppletCtl *ctl;
    AtkObject *atk;
    
    _F();

    if (instances==1)
        vm_init_icons();

    panel_applet_set_flags(applet, PANEL_APPLET_EXPAND_MINOR); 

    ctl = vm_ctl_new(applet);
    
        
    panel_applet_setup_menu_from_file ( PANEL_APPLET (applet), 
      	        		                DATADIR,
                                        "GNOME_VmApplet.xml",
                                        NULL,
                                        vm_menu_verbs,
                                        ctl);

    if (panel_applet_get_locked_down (PANEL_APPLET (applet))) 
    {
        BonoboUIComponent *popup_component;

	    popup_component = panel_applet_get_popup_component (PANEL_APPLET (applet));
        bonobo_ui_component_set_prop (popup_component,
					"/commands/preferences",
					"hidden", "1",
					NULL);
    }

	g_signal_connect(G_OBJECT(ctl->applet), "button-press-event",
			G_CALLBACK(vm_button_cb), ctl);
    g_signal_connect(G_OBJECT(ctl->applet), "destroy",
			G_CALLBACK(vm_destroy_cb), ctl);
    g_signal_connect(G_OBJECT(ctl->applet), "size_allocate",
   		    G_CALLBACK(vm_size_cb), ctl);
    
    atk = gtk_widget_get_accessible (ctl->applet);
    if (GTK_IS_ACCESSIBLE (atk)) 
    {
        atk_object_set_name (atk, _("VmApplet"));
        atk_object_set_description(atk, _("Monitor locally running virtual domains"));
    }    
    
    gtk_widget_show_all(ctl->applet);
    
    ctl->applet_tip =  gtk_tooltips_new();
   
    vm_update(ctl);

    vm_icon_update(ctl, panel_applet_get_size(PANEL_APPLET(ctl->applet)), FALSE);
    
    vm_timer_refresh(ctl); 
       
    return TRUE;
}

/*
 * "main()" of the applet
 */
static gboolean
vm_factory(PanelApplet *applet, const gchar *iid, gpointer ptr)
{
	gboolean retval = FALSE;
    
    instances++;
    _D("---- INITIALIZE instance: %d ----", instances);
    
    _F();
    
    if (!strcmp(iid, "OAFIID:GNOME_VmApplet")) 
        retval = vm_setup(applet);

    return retval;
}

/*
 * Bonobo initialization
 */
PANEL_APPLET_BONOBO_FACTORY("OAFIID:GNOME_VmApplet_Factory",
                             PANEL_TYPE_APPLET,
                             "vm-applet",
                             "0",
                             (PanelAppletFactoryCallback) vm_factory,
                             NULL)


/*
 * vim: set tabstop=4:
 * vim: set shiftwidth=4:
 * vim: set expandtab:
 */
