/*
 * =====================================================================================
 * 
 *        Filename:  applet.h
 * 
 *         Version:  $Id: applet.h,v 1.6 2006/07/13 23:53:48 kzak Exp $
 *         Created:  15.12.2005 20:22:13 CET
 * 
 *          Author:  Karel Zak <kzak@redhat.com>
 * 
 * =====================================================================================
 */

#ifndef __APPLET_H__
#define __APPLET_H__

#include <gdk-pixbuf/gdk-pixbuf.h>
#include <gtk/gtksizegroup.h>
#include <gtk/gtktooltips.h>
#include <gtk/gtkwidget.h>
#include <panel-applet.h>

#include <libvirt/libvirt.h>

/*#define DEBUG   1*/

#ifdef DEBUG
#define _D(_format, _a...)      fprintf(stderr, _format "\n", ##_a)
#define _F()                    fprintf(stderr, "--> %s()\n", __FUNCTION__)
#else
#define _D(_format, _a...)
#define _F()
#endif

/* dummy */
typedef struct _VmAppletCtl VmAppletCtl;
typedef struct _VmAppletDomain VmAppletDomain;

/*
 * Name of external wrapper
 */
#define VM_WRAPPER_NAME		"vm_applet_wrapper"

/*
 * Icon stock names
 */
#define VM_STOCK_NONE		NULL

#define VM_STOCK_BLOCKED	"vm-stock-blocked"
#define VM_STOCK_PAUSED		"vm-stock-paused"
#define VM_STOCK_RUNNING	"vm-stock-running"
#define VM_STOCK_SHUTDOWN	"vm-stock-shutdown"
#define VM_STOCK_SHUTOFF	"vm-stock-shutoff"
#define VM_STOCK_UNKNOWN	"vm-stock-unknonw"

#define VM_STOCK_DO_TERMINAL	"vm-stock-do-terminal"
#define VM_STOCK_DO_DESTROY	"vm-stock-do-destroy"
#define VM_STOCK_DO_RESUME	"vm-stock-do-resume"
#define VM_STOCK_DO_SHUTDOWN	"vm-stock-do-shutdown"
#define VM_STOCK_DO_SUSPEND	"vm-stock-do-suspend"

#define VM_STOCK_DO_DETAILS     GTK_STOCK_JUMP_TO
#define VM_STOCK_DO_CONSOLE     "vm-stock-do-console"
#define VM_STOCK_DO_EDIT        GTK_STOCK_PROPERTIES

/*
 * GConf stuff
 */
#define VM_GCONF_TERMINAL_CMD		"terminal_cmd"
#define VM_GCONF_TERMINAL_ENABLE	"terminal_enb"
#define VM_GCONF_DELAY			"delay"

/*
 * Icon description
 */
typedef struct
{
	gchar	*stock_id;
	gchar	*filename;
} VmAppletStockIcon;

/* the domain without CPU time change for VM_BLOCKED_TIME 
 * (milliseconds) is marked as "blocked"
 */
#define VM_BLOCKED_TIME		10

/* misc */
#define VM_UUID_SIZ		37	/* bytes */
#define VM_DEFAULT_URI		"Xen"

typedef struct VmAppletConn
{
	virConnectPtr		ptr;		/* libvirt connection */
	char			*uri;		/* URI */
	gboolean		readonly;
	VmAppletDomain		*dom;		/* Vm domains */
	int			ndoms;		/* number of Vm domains */
	virNodeInfo		ninfo;		/* this node HW info */	
	struct VmAppletConn	*next;
} VmAppletConn;

/* 
 * Domain description 
 */
typedef struct _VmAppletDomain
{
	int			id;		/* domain ID */
	char			*name;		/* domain Name */
	char			uuid[VM_UUID_SIZ]; /* UUID */
	char			*ostype;	/* domain OS Type */
	virDomainState		state;		/* domain State */
	virDomainState		activity;	/* real state by State + CpuTime */
	unsigned long		maxmem;		/* domain Max allocated Memory */
	unsigned long		usemem;		/* domain Used Memory */
	unsigned long long	cputime;	/* domain CPU Time */
	int			cputime_ct;	/* number of loops without CPU Time change */
	int 			ncpus;		/* domain number of CPUs */
	GtkWidget		*menuitem;	/* popup menu item */
	VmAppletCtl		*ctl;		/* global applet control struct */
	int			changed;	/* bit area with info about changed columns */
	VmAppletConn		*conn;		/* paretn */
	VmAppletDomain		*next;		/* next domain */
} _VmAppletDomain;

/*
 * Domain data IDs
 */
#define VM_DATA_ID       (1 << 1)
#define VM_DATA_NAME     (1 << 2)
#define VM_DATA_OSTYPE   (1 << 3)
#define VM_DATA_ACTIVITY (1 << 4)
#define VM_DATA_MAXMEM   (1 << 5)
#define VM_DATA_USEDMEM	 (1 << 6)
#define VM_DATA_NCPUS    (1 << 7)
#define VM_DATA_CPUTIME  (1 << 8)
#define VM_DATA_STATE    (1 << 9)
#define VM_DATA_UUID	 (1 << 10)

/* 
 * Domain menuitem description 
 */
typedef struct VmAppletMenuDesc
{
	int			did;		/* data source ID */
	char			*name;		/* column name */
	char			*gconf;		/* pointer to gnome config */
	GtkWidget		*(*builder)(	VmAppletDomain *, 
						struct VmAppletMenuDesc *, 
						GtkWidget *);
	gboolean		enable;			/* visible? */
	GtkSizeGroup		*sizegroup;	/* GTK size group */
	VmAppletCtl		*ctl;		/* global applet control struct */
} VmAppletMenuDesc;

/* 
 * Internal applet control struct 
 */
typedef struct _VmAppletCtl
{
	GtkWidget		*applet;	/* applet widget */
	GtkTooltips		*applet_tip;	/* applet tip */
	GtkWidget		*image;		/* applet icon */
	GdkPixbuf 		*pixbuf;	/* icon pixbuf */
	int			image_size;	/* last used applet icon size */
	GtkWidget		*popmenu;	/* list of domais */
	GtkWidget		*propwin;	/* preferences window */
	GtkWidget		*actmenu;	/* actions menu */
	int			popmenu_up;	/* is the pop menu UP/DOWN */
	VmAppletMenuDesc	*desc;		/* menu columns descriptions */
	int			ndescs;		/* num of items in "descs" */
	VmAppletDomain		*actdom;	/* selected domain */
	guint  			pixtimer;	/* glib timer */
	int 			delay;		/* update delay */
	uid_t			uid;		/* who runs me */
	int			instance;	/* ID of applet instance */
	int			ndoms;		/* number of domains */
	char			*terminal_cmd;	/* used defined terminal command */
	gboolean		terminal_enb;	/* To be, or not to be? That is the question! */
	VmAppletConn		*conn;		/* List of connections */	
	int			active_conns;	/* number of active connections */
} _VmAppletCtl; 


/*
 * Domain status masks
 */
#define vm_state2mask(_s)	(1 << ((int)_s))

typedef enum {
	VM_SMASK_NONE		= 0,
	VM_SMASK_RUNNING	= vm_state2mask(VIR_DOMAIN_RUNNING),
	VM_SMASK_BLOCKED	= vm_state2mask(VIR_DOMAIN_BLOCKED),
	VM_SMASK_PAUSED		= vm_state2mask(VIR_DOMAIN_PAUSED),
	VM_SMASK_SHUTDOWN	= vm_state2mask(VIR_DOMAIN_SHUTDOWN),
	VM_SMASK_SHUTOFF	= vm_state2mask(VIR_DOMAIN_SHUTOFF),

	VM_SMASK_ALL		= (VM_SMASK_RUNNING | VM_SMASK_BLOCKED | 
				   VM_SMASK_PAUSED | VM_SMASK_SHUTDOWN | VM_SMASK_SHUTOFF )
} VmAppetSmask;

#define vm_state_is_valid(_s)	(vm_state2mask(_s) & VM_SMASK_ALL)

/*
 * Action menu rows
 */
typedef struct VmAppletActmenuRow
{
	char		*label;
	void		(*callback)(GtkMenuItem *, gpointer);
	const char	*stock_id;
	int		dom0;		/* enable this itemo for dom0? */
	VmAppetSmask	mask;
} VmAppletActmenuRow;

/* 
 * D-BUS
 */
#define VM_DBUS_VIRTMANAGER_SERVICE	"com.redhat.virt.manager"
#define VM_DBUS_VIRTMANAGER_PATH	"/com/redhat/virt/manager"
#define VM_DBUS_VIRTMANAGER_INTERFACE	"com.redhat.virt.manager"

/*
 * Exported routines
 */
extern void vm_show_help( VmAppletCtl *ctl, const char *section );
extern void vm_prop_show(VmAppletCtl *ctl);
extern void vm_timer_refresh(VmAppletCtl *ctl);
extern void vm_popmenu_destroy_items(VmAppletCtl *ctl);

#endif /* __APPLET_H__ */

