/*
 * =====================================================================================
 * 
 *        Filename:  wrapper.c
 * 
 *     Description:  Consolehelper wrapper
 * 
 *         Version:  $Id: wrapper.c,v 1.2 2006/07/13 13:39:36 kzak Exp $
 *         Created:  10.02.2006 10:56:18 CET
 * 
 *          Author:  Karel Zak <kzak@redhat.com>
 * 
 * =====================================================================================
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <sys/types.h>
#include <libvirt/libvirt.h>

typedef struct VmWrapperCmd
{
	char	*name;
	int	(*doit)(virConnectPtr conn, virDomainPtr dom, void *data);
} VmWrapperCmd;

static int cmd_shutdown(virConnectPtr conn, virDomainPtr dom, void *data);
static int cmd_destroy(virConnectPtr conn, virDomainPtr dom, void *data);
static int cmd_suspend(virConnectPtr conn, virDomainPtr dom, void *data);
static int cmd_resume(virConnectPtr conn, virDomainPtr dom, void *data);
static int cmd_terminal(virConnectPtr conn, virDomainPtr dom, void *data);

VmWrapperCmd commands[] = 
{
	{ "shutdown",	cmd_shutdown },
	{ "destroy",	cmd_destroy },
	{ "suspend",	cmd_suspend },	/* pause */
	{ "resume",	cmd_resume },	/* unpause */
	{ "terminal",	cmd_terminal },
/**** TODO:
	{ "save",	cmd_save },	
	{ "restore",	cmd_restore },
	{ "setmaxmem",	cmd_setmaxmem },
****/
	{ NULL, NULL }
};

static int
cmd_shutdown(virConnectPtr conn, virDomainPtr dom, void *data)
{
	return virDomainShutdown(dom)==0 ? 1 : 0;
}
static int
cmd_destroy(virConnectPtr conn, virDomainPtr dom, void *data)
{
	return virDomainDestroy(dom)==0 ? 1 : 0;
}
static int
cmd_suspend(virConnectPtr conn, virDomainPtr dom, void *data)
{
	return virDomainSuspend(dom)==0 ? 1 : 0;
}
static int
cmd_resume(virConnectPtr conn, virDomainPtr dom, void *data)
{
	return virDomainResume(dom)==0 ? 1 : 0;
}
static int
cmd_terminal(virConnectPtr conn, virDomainPtr dom, void *data)
{
	unsigned int id = virDomainGetID(dom); 
	char idstr[64];
	char *arg0;
		
	if (dom)
		virDomainFree(dom);
	 
	virConnectClose(conn);

	/* argv[0] */
	if ((arg0 = strrchr(VM_XENTERMINAL, '/')))
		++arg0;
	else
		arg0 = VM_XENTERMINAL;
	
	/* argv[1] */
	snprintf(idstr, sizeof(idstr), "%u", id);
	
	if (execl(VM_XENTERMINAL, arg0, idstr, NULL)==-1)
	{
		perror(VM_XENTERMINAL);
		exit(EXIT_FAILURE);
	}
	
	exit(EXIT_SUCCESS);
}

int
main(int argc, char **argv)
{
	virDomainPtr domain;
	VmWrapperCmd *cmd = NULL, *cp;
	virConnectPtr conn;
	char *cmdstr, *uuid;
	int ret;
	
	if (getuid()!=0)
	{
		fprintf(stderr, "%s: I work for root user only! Who are you?\n", argv[0]);
		exit(EXIT_FAILURE);
	}
	if (argc < 3)
	{
		fprintf(stderr, "Usage: %s <command> <domainUUID>\ncommands:\n", argv[0]);
		for (cp=commands; cp->name; cp++)
			fprintf(stderr, "   %s\n", cp->name);
		exit(EXIT_FAILURE);
	}

	cmdstr = argv[1];
	uuid = argv[2];

	if (cmdstr==NULL || *cmdstr=='\0' || uuid==NULL)
		exit(EXIT_FAILURE);

	for (cp=commands; cp->name; cp++)
	{
		if (strcasecmp(cp->name, cmdstr)==0)
		{
			cmd = cp;
			break;
		}
	}
	if (!cmd)
	{
		fprintf(stderr, "%s: unsupported command '%s'\n", argv[0], cmdstr);
		exit(EXIT_FAILURE);
	}
	if (!(conn = virConnectOpen(NULL)))
	{
		fprintf(stderr, "%s: failed to connect to the hypervisor\n", argv[0]);
		exit(EXIT_FAILURE);
	}
	
	
	if (!(domain = virDomainLookupByUUIDString(conn, uuid)))
	{
		fprintf(stderr, "%s: failed to get domain '%s'\n", argv[0], uuid);
		ret = EXIT_FAILURE;
	}
	else
		ret = cmd->doit(conn, domain, NULL) ? EXIT_SUCCESS : EXIT_FAILURE;
	
	if (domain)
		virDomainFree(domain);
	 
	virConnectClose(conn);

	exit(ret);
}

